@@ -12,7 +12,7 @@ use feature ();
 package Zydeco;
 
 our $AUTHORITY = 'cpan:TOBYINK';
-our $VERSION   = '0.520';
+our $VERSION   = '0.521';
 
 use Keyword::Simple ();
 use PPR;
@@ -1018,17 +1018,33 @@ sub _handle_name_list {
 	return @names;
 }
 
+my $_should_optimize = sub {
+	my ($code, $sigvars) = @_;
+	
+	my %allowed = ( '$self' => undef, '$class' => undef, '$_' => undef, '@_' => undef );
+	undef $allowed{$_} for split /\s*,\s*/, $sigvars;
+	
+	my @vars = ( $code =~ /[\$\@\%]\w+/g );
+	foreach my $var (@vars) {
+		next if exists $allowed{$var};
+		return 0;
+	}
+	1;
+};
+
 sub _handle_factory_keyword {
 	my ($me, $name, $via, $code, $has_sig, $sig, $attrs) = @_;
 	
+	my ($signature_is_named, $signature_var_list, $type_params_stuff, $extra) = $has_sig ? $me->_handle_signature_list($sig) : ();
+	
 	my $optim;
 	for my $attr (@$attrs) {
 		$optim = 1 if $attr =~ /^:optimize\b/;
 	}
 	
 	if (defined $code and $code =~ /^=(.+)$/s) {
-		$code  = "{ $1 }";
-		$optim = 1;
+		$code    = "{ $1 }";
+		$optim ||= $_should_optimize->($code, $signature_var_list);
 	}
 	
 	if ($via) {
@@ -1050,7 +1066,6 @@ sub _handle_factory_keyword {
 			!!$optim,
 		);
 	}
-	my ($signature_is_named, $signature_var_list, $type_params_stuff, $extra) = $me->_handle_signature_list($sig);
 	my $munged_code = sprintf('sub { my($factory,$class,%s)=(shift,shift,@_); %s; do %s }', $signature_var_list, $extra, $code);
 	sprintf(
 		'q[%s]->_factory(%s, { attributes => %s, caller => __PACKAGE__, code => %s, named => %d, signature => %s, optimize => %d });',
@@ -1068,14 +1083,16 @@ sub _handle_method_keyword {
 	my $me = shift;
 	my ($name, $code, $has_sig, $sig, $attrs) = @_;
 
+	my ($signature_is_named, $signature_var_list, $type_params_stuff, $extra) = $has_sig ? $me->_handle_signature_list($sig) : ();
+
 	my $optim;
 	for my $attr (@$attrs) {
 		$optim = 1 if $attr =~ /^:optimize\b/;
 	}
 	
 	if (defined $code and $code =~ /^=(.+)$/s) {
-		$code  = "{ $1 }";
-		$optim = 1;
+		$code    = "{ $1 }";
+		$optim ||= $_should_optimize->($code, $signature_var_list);
 	}
 	
 	my $lex_name;
@@ -1087,7 +1104,6 @@ sub _handle_method_keyword {
 	
 	if (defined $name and not defined $lex_name) {
 		if ($has_sig) {
-			my ($signature_is_named, $signature_var_list, $type_params_stuff, $extra) = $me->_handle_signature_list($sig);
 			my $munged_code = sprintf('sub { my($self,%s)=(shift,@_); %s; my $class = ref($self)||$self; do %s }', $signature_var_list, $extra, $code);
 			$return = sprintf(
 				'q[%s]->_can(%s, { attributes => %s, caller => __PACKAGE__, code => %s, named => %d, signature => %s, optimize => %d });',
@@ -1114,7 +1130,6 @@ sub _handle_method_keyword {
 	}
 	else {
 		if ($has_sig) {
-			my ($signature_is_named, $signature_var_list, $type_params_stuff, $extra) = $me->_handle_signature_list($sig);
 			my $munged_code = sprintf('sub { my($self,%s)=(shift,@_); %s; my $class = ref($self)||$self; do %s }', $signature_var_list, $extra, $code);
 			$return = sprintf(
 				'q[%s]->wrap_coderef({ attributes => %s, caller => __PACKAGE__, code => %s, named => %d, signature => %s, optimize => %d });',
@@ -1149,6 +1164,8 @@ sub _handle_multi_keyword {
 	my $me = shift;
 	my ($kind, $name, $code, $has_sig, $sig, $attrs) = @_;
 	
+	my ($signature_is_named, $signature_var_list, $type_params_stuff, $extra) = $has_sig ? $me->_handle_signature_list($sig) : ();
+	
 	my $optim;
 	my $extra_code = '';
 	for my $attr (@$attrs) {
@@ -1159,12 +1176,11 @@ sub _handle_multi_keyword {
 	}
 	
 	if (defined $code and $code =~ /^=(.+)$/s) {
-		$code  = "{ $1 }";
-		$optim = 1;
+		$code    = "{ $1 }";
+		$optim ||= $_should_optimize->($code, $signature_var_list);
 	}
 	
 	if ($has_sig) {
-		my ($signature_is_named, $signature_var_list, $type_params_stuff, $extra) = $me->_handle_signature_list($sig);
 		my $munged_code = sprintf('sub { my($self,%s)=(shift,@_); %s; my $class = ref($self)||$self; do %s }', $signature_var_list, $extra, $code);
 		return sprintf(
 			'q[%s]->_multi(%s => %s, { attributes => %s, caller => __PACKAGE__, code => %s, named => %d, signature => %s, %s });',
@@ -1195,14 +1211,16 @@ sub _handle_multi_keyword {
 sub _handle_modifier_keyword {
 	my ($me, $kind, $names, $code, $has_sig, $sig, $attrs) = @_;
 
+	my ($signature_is_named, $signature_var_list, $type_params_stuff, $extra) = $has_sig ? $me->_handle_signature_list($sig) : ();
+	
 	my $optim;
 	for my $attr (@$attrs) {
 		$optim = 1 if $attr =~ /^:optimize\b/;
 	}
 	
 	if (defined $code and $code =~ /^=(.+)$/s) {
-		$code  = "{ $1 }";
-		$optim = 1;
+		$code    = "{ $1 }";
+		$optim ||= $_should_optimize->($code, $signature_var_list);
 	}
 	
 	# MooX::Press cannot handle optimizing method modifiers
@@ -1215,7 +1233,6 @@ sub _handle_modifier_keyword {
 		map { /^\{/ ? "scalar(do $_)" : B::perlstring($_) } @names;
 
 	if ($has_sig) {
-		my ($signature_is_named, $signature_var_list, $type_params_stuff, $extra) = $me->_handle_signature_list($sig);
 		my $munged_code;
 		if ($kind eq 'around') {
 			$munged_code = sprintf('sub { my($next,$self,%s)=(shift,shift,@_); %s; my $class = ref($self)||$self; do %s }', $signature_var_list, $extra, $code);
@@ -2322,7 +2339,7 @@ sub _include {
 #{
 #	package Zydeco::Anonymous::Package;
 #	our $AUTHORITY = 'cpan:TOBYINK';
-#	our $VERSION   = '0.520';
+#	our $VERSION   = '0.521';
 #	use overload q[""] => sub { ${$_[0]} }, fallback => 1;
 #	sub DESTROY {}
 #	sub AUTOLOAD {
@@ -2333,7 +2350,7 @@ sub _include {
 #	
 #	package Zydeco::Anonymous::Class;
 #	our $AUTHORITY = 'cpan:TOBYINK';
-#	our $VERSION   = '0.520';
+#	our $VERSION   = '0.521';
 #	our @ISA       = qw(Zydeco::Anonymous::Package);
 #	sub new {
 #		my $me = shift;
@@ -2346,12 +2363,12 @@ sub _include {
 #	
 #	package Zydeco::Anonymous::Role;
 #	our $AUTHORITY = 'cpan:TOBYINK';
-#	our $VERSION   = '0.520';
+#	our $VERSION   = '0.521';
 #	our @ISA       = qw(Zydeco::Anonymous::Package);
 #	
 #	package Zydeco::Anonymous::ParameterizableClass;
 #	our $AUTHORITY = 'cpan:TOBYINK';
-#	our $VERSION   = '0.520';
+#	our $VERSION   = '0.521';
 #	our @ISA       = qw(Zydeco::Anonymous::Package);
 #	sub generate_package {
 #		my $me  = shift;
@@ -2365,7 +2382,7 @@ sub _include {
 #
 #	package Zydeco::Anonymous::ParameterizableRole;
 #	our $AUTHORITY = 'cpan:TOBYINK';
-#	our $VERSION   = '0.520';
+#	our $VERSION   = '0.521';
 #	our @ISA       = qw(Zydeco::Anonymous::Package);
 #	sub generate_package {
 #		my $me  = shift;
